import copy
import re
import inspect
from functools import partial
from collections import namedtuple

import numpy

__version__ = '0.2.0'


def decode_IDN(s):
    manuf, model, serial, version = map(str.strip, s.split(","))
    return dict(manufacturer=manuf, model=model, serial=serial, version=version)


def __decode_Err(s):
    code, desc = map(str.strip, s.split(",", 1))
    return dict(code=int(code), desc=desc[1:-1])


def __decode_ErrArray(s):
    msgs = list(map(str.strip, s.split(",")))
    result = []
    for i in range(0, len(msgs), 2):
        code, desc = int(msgs[i]), msgs[i + 1][1:-1]
        if code == 0:
            continue
        result.append(dict(code=code, desc=desc))
    return result


def __decode_OnOff(s):
    su = s.upper()
    if su in ("1", "ON"):
        return True
    elif su in ("0", "OFF"):
        return False
    else:
        raise ValueError("Cannot decode OnOff value {0}".format(s))


def __encode_OnOff(s):
    if s in (0, False, "off", "OFF"):
        return "OFF"
    elif s in (1, True, "on", "ON"):
        return "ON"
    else:
        raise ValueError("Cannot encode OnOff value {0}".format(s))


__decode_IntArray = partial(numpy.fromstring, dtype=int, sep=",")
__decode_FloatArray = partial(numpy.fromstring, dtype=float, sep=",")

#: SCPI command
#: accepts the following keys:
#:
#:   - get - translation function called on the result of a query.
#:           If not present means command cannot be queried.
#:           If present and is None means ignore query result
#:   - set - translation function called before a write.
#:           If not present means command cannot be written.
#:           If present and is None means it doesn't receive any argument
#:   - doc - command documentation (str, optional)
Cmd = dict

FuncCmd = partial(Cmd, set=None)

IntCmd = partial(Cmd, get=int, set=str)
IntCmdR = partial(Cmd, get=int)
IntCmdW = partial(Cmd, set=str)

FloatCmd = partial(Cmd, get=float, set=str)
FloatCmdR = partial(Cmd, get=float)
FloatCmdW = partial(Cmd, set=str)

StrCmd = partial(Cmd, get=str, set=str)
StrCmdR = partial(Cmd, get=str)
StrCmdW = partial(Cmd, set=str)

IntArrayCmdR = partial(Cmd, get=__decode_IntArray)
FloatArrayCmdR = partial(Cmd, get=__decode_FloatArray)
StrArrayCmd = partial(Cmd, get=lambda x: x.split(","), set=lambda x: ",".join(x))
StrArrayCmdR = partial(Cmd, get=lambda x: x.split(","))

OnOffCmd = partial(Cmd, get=__decode_OnOff, set=__encode_OnOff)
OnOffCmdR = partial(Cmd, get=__decode_OnOff)
OnOffCmdW = partial(Cmd, set=__encode_OnOff)
BoolCmd = OnOffCmd
BoolCmdR = OnOffCmdR
BoolCmdW = OnOffCmdW

IDNCmd = partial(Cmd, get=decode_IDN, doc="identification query")

ErrCmd = partial(Cmd, get=__decode_Err)
ErrArrayCmd = partial(Cmd, get=__decode_ErrArray)


def min_max_cmd(cmd_expr):
    """
    Find the shortest and longest version of an SCPI command expression.

    Example:
        >>> min_max_cmd('SYSTem:ERRor[:NEXT]')
        ('SYST:ERR', 'SYSTEM:ERROR:NEXT')
    """
    result_min, optional = "", 0
    for c in cmd_expr:
        if c.islower():
            continue
        if c == "[":
            optional += 1
            continue
        if c == "]":
            optional -= 1
            continue
        if optional:
            continue
        result_min += c
    result_min = result_min.lstrip(":")
    result_max = cmd_expr.replace("[", "").replace("]", "").upper().lstrip(":")
    return result_min, result_max


def cmd_expr_to_reg_expr_str(cmd_expr):
    """
    Return a regular expression string from the given SCPI command expression.
    """
    # Basicaly we replace [] -> ()?, and LOWercase -> LOW(ercase)?
    # Also we add :? optional to the start and $ to the end to make sure
    # we have an exact match
    reg_expr, low_zone = r"\:?", False
    for c in cmd_expr:
        cl = c.islower()
        if not cl:
            if low_zone:
                reg_expr += ")?"
            low_zone = False
        if c == "[":
            reg_expr += "("
        elif c == "]":
            reg_expr += ")?"
        elif cl:
            if not low_zone:
                reg_expr += "("
            low_zone = True
            reg_expr += c.upper()
        elif c in "*:":
            reg_expr += "\\" + c
        else:
            reg_expr += c

    # if cmd expr ends in lower case we close the optional zone 'by hand'
    if low_zone:
        reg_expr += ")?"

    return reg_expr + "$"


def cmd_expr_to_reg_expr(cmd_expr):
    """
    Return a compiled regular expression object from the given SCPI command
    expression.
    """
    return re.compile(cmd_expr_to_reg_expr_str(cmd_expr), re.IGNORECASE)


class Commands:
    """
    A dict like container for SCPI commands. Construct a Commands object like a
    dict.  When creating a Commands object, *args* must either:

    * another *Commands* object
    * a dict where keys should be SCPI command expressions
      (ex: `SYSTem:ERRor[:NEXT]`). Values can be any python objects
      Common use-case is instance of *Cmd* which contain information on how to
      handle a get/set operation and optional docstrings
    * a sequence of pairs where first element must be SCPI command expression
      and second element is any python object

    *kwargs* should also be SCPI command expressions;

    The same way, assignment keys should be SCPI command expressions and
    assignment values can be any python object.

    Examples::

        from egse.scpi import FuncCmd, ErrCmd, IntCmd, Commands

        # c1 will only have *CLS command
        c1 = Commands({'*CLS': FuncCmd(doc='clear status'),
                       '*RST': FuncCmd(doc='reset')})

        # c2 will have *CLS and VOLTage commands
        c2 = Commands(c1, VOLTage=IntCmd())

        # add error command to c2
        c2['SYSTem:ERRor[:NEXT]'] = ErrCmd()

    Access to a command will return the same python object for any text
    that resolves to the same SCPI command::

        >>> c2['SYST:ERR']
        <ErrCmd at 0x7f3f663bf390>
        >>> c2[':system:error:next']
        <ErrCmd at 0x7f3f663bf390>
        >>> c2[':system:error:next'] == c2['SYST:ERR']
        True
    """

    def __init__(self, *args, **kwargs):
        self.command_expressions = {}
        self._command_cache = {}
        for arg in args:
            self.update(arg)
        self.update(kwargs)

    def __setitem__(self, cmd_expr, command):
        min_cmd, max_cmd = min_max_cmd(cmd_expr)
        cmd_info = dict(
            value=command,
            re=cmd_expr_to_reg_expr(cmd_expr),
            min_command=min_cmd,
            max_command=max_cmd,
        )
        self.command_expressions[cmd_expr] = cmd_info
        # update cache with short and long version
        self.get_command_expression(min_cmd)
        self.get_command_expression(max_cmd)
        return cmd_info

    def __getitem__(self, cmd_name):
        return self.get_command(cmd_name)['value']

    def __delitem__(self, cmd_expr):
        reg_expr = self.command_expressions.pop(cmd_expr)['re']
        del_cache = {k for k, v in self._command_cache.items()
                     if reg_expr.match(k)}
        for k in del_cache:
            del self._command_cache[k]

    def __contains__(self, cmd_name):
        return self.get(cmd_name) is not None

    def __iter__(self):
        return self.command_expressions.__iter__()

    def __len__(self):
        return len(self.command_expressions)

    def deepcopy(self):
        return copy.deepcopy(self.command_expressions)

    def clear(self):
        self.command_expressions.clear()
        self._command_cache.clear()

    def keys(self):
        return self.command_expressions.keys()

    def values(self):
        return {k:v['value'] for k, v in self.command_expressions.items()}.values()

    def get_command(self, cmd_name):
        cmd_expr = self.get_command_expression(cmd_name)
        return self.command_expressions[cmd_expr]

    def get_command_expression(self, cmd_name):
        cmd_name_u = cmd_name.upper()
        try:
            return self._command_cache[cmd_name_u]
        except KeyError:
            for cmd_expr, cmd_info in self.command_expressions.items():
                reg_expr = cmd_info["re"]
                if reg_expr.match(cmd_name):
                    self._command_cache[cmd_name.upper()] = cmd_expr
                    return cmd_expr
        raise KeyError(cmd_name)

    def get(self, cmd_name, default=None):
        try:
            return self[cmd_name]
        except KeyError:
            return default

    def update(self, commands):
        if isinstance(commands, Commands):
            self.command_expressions.update(commands.command_expressions)
            self._command_cache.update(commands._command_cache)
        elif isinstance(commands, dict):
            for cmd_expr, cmd in commands.items():
                self[cmd_expr] = cmd
        else:
            for cmd_expr, cmd in commands:
                self[cmd_expr] = cmd


COMMANDS = Commands(
    {
        "*CLS": FuncCmd(doc="clear status"),
        "*ESE": IntCmd(doc="standard event status enable register"),
        "*ESR": IntCmdR(doc="standard event event status register"),
        "*IDN": IDNCmd(),
        "*OPC": IntCmdR(set=None, doc="operation complete"),
        "*OPT": IntCmdR(doc="return model number of any installed options"),
        "*RCL": IntCmdW(set=int, doc="return to user saved setup"),
        "*RST": FuncCmd(doc="reset"),
        "*SAV": IntCmdW(doc="save the preset setup as the user-saved setup"),
        "*SRE": IntCmdW(doc="service request enable register"),
        "*STB": StrCmdR(doc="status byte register"),
        "*TRG": FuncCmd(doc="bus trigger"),
        "*TST": Cmd(get=lambda x: not __decode_OnOff(x), doc="self-test query"),
        "*WAI": FuncCmd(doc="wait to continue"),
        "SYSTem:ERRor[:NEXT]": ErrCmd(doc="return and clear oldest system error"),
    }
)


class SCPIError(Exception):
    """
    Base :term:`SCPI` error
    """


def sanitize_msgs(*msgs, eol='\n', sep=';', strict_query=True):
    """
    Transform a tuple of messages into a list  of
    (<individual commands>, <individual queries>, <full_message>):

    if strict_query=True, sep=';', eol='\n' (default):
        msgs = ('*rst', '*idn?;*cls') =>
            (['*RST', '*IDN?', '*CLS'], ['*IDN?'], '*RST\n*IDN?\n*CLS')

    if strict_query=False, sep=';', eol='\n' (default):
        msgs = ('*rst', '*idn?;*cls') =>
            (['*RST', '*IDN?', '*CLS'], ['*IDN?'], '*RST\n*IDN?;*CLS')
    """
    # in case a single message comes with several eol separated commands
    msgs = eol.join(msgs).split(eol)
    result, commands, queries = [], [], []
    for msg in msgs:
        sub_result = []
        for cmd in msg.split(sep):
            cmd = cmd.strip()
            if not cmd:
                continue
            commands.append(cmd)
            is_query = "?" in cmd
            if is_query:
                queries.append(cmd)
            if is_query and strict_query:
                if sub_result:
                    result.append(sep.join(sub_result))
                    sub_result = []
                result.append(cmd)
            else:
                sub_result.append(cmd)
        if sub_result:
            result.append(sep.join(sub_result))
    return commands, queries, eol.join(result) + eol


Request = namedtuple('Request', 'name args query')


def split_line(line, sep=';'):
    line = line.strip().strip(';')
    requests = []
    for req_str in line.split(sep):
        if not req_str:
            continue
        query = '?' in req_str
        req_str = req_str.replace('?', '')
        name, _, args = req_str.partition(' ')
        name, args = name.strip(), args.strip()
        requests.append(Request(name, args, query))
    return requests



"""
The control server for the dummy device.
"""
import logging
import multiprocessing
import random
import sys
import threading
import time
from functools import partial

import typer
import zmq
from egse.command import ClientServerCommand
from egse.confman import is_configuration_manager_active
from egse.control import ControlServer
from egse.control import is_control_server_active
from egse.decorators import dynamic_interface
from egse.listener import Event
from egse.listener import EventInterface
from egse.protocol import CommandProtocol
from egse.proxy import Proxy
from egse.settings import Settings
from egse.storage import TYPES
from egse.storage import is_storage_manager_active
from egse.storage import register_to_storage_manager
from egse.storage import store_housekeeping_information
from egse.storage import unregister_from_storage_manager
from egse.system import attrdict
from egse.system import format_datetime
from egse.zmq_ser import bind_address
from egse.zmq_ser import connect_address

from cgse_dummy.dummy_devif import DummyEthernetInterface

_LOGGER = logging.getLogger("cgse_dummy.dummy.cs")

cs_settings = Settings.load("DUMMY CS")
dev_settings = Settings.load("DUMMY DEVICE")

DEV_HOST = "localhost"
"""The hostname or IP address of the Dummy Device."""
DEV_PORT = dev_settings.PORT
"""The port number for the Dummy Device."""
DEV_NAME = f"Dummy Device {dev_settings.MODEL}"
"""The name used for the Dummy Device, this is used in Exceptions and in the info command."""

READ_TIMEOUT = 10.0  # seconds
"""The maximum time to wait for a socket receive command."""
WRITE_TIMEOUT = 1.0  # seconds
"""The maximum time to wait for a socket send command."""
CONNECT_TIMEOUT = 3.0  # seconds
"""The maximum time to wait for establishing a socket connect."""

# Especially DummyCommand and DummyController need to be defined in a known module
# because those objects are pickled and when de-pickled at the clients side the class
# definition must be known.

commands = attrdict(
    {
        'info': {
            'description': 'Info on the Dummy Device.',
            'response': 'handle_device_method'
        },
        'get_value': {
            'description': 'Read a value from the device.',
        },
        'handle_event': {
            'description': "Notification of an event",
            'device_method': 'handle_event',
            'cmd': '{event}',
            'response': 'handle_device_method'
        },
    }
)

app = typer.Typer(help=f"Dummy control server for the dummy device {dev_settings.MODEL}")


def is_dummy_cs_active():
    return is_control_server_active(
        endpoint=connect_address(cs_settings.PROTOCOL, cs_settings.HOSTNAME, cs_settings.COMMANDING_PORT)
    )


class DummyCommand(ClientServerCommand):
    pass


class DummyInterface:
    @dynamic_interface
    def info(self):
        ...

    @dynamic_interface
    def get_value(self, *args, **kwargs):
        ...


class DummyProxy(Proxy, DummyInterface, EventInterface):
    def __init__(
            self,
            protocol=cs_settings.PROTOCOL,
            hostname=cs_settings.HOSTNAME,
            port=cs_settings.COMMANDING_PORT,
            timeout=cs_settings.TIMEOUT
    ):
        """
        Args:
            protocol: the transport protocol [default is taken from settings file]
            hostname: location of the control server (IP address) [default is taken from settings file]
            port: TCP port on which the control server is listening for commands [default is taken from settings file]
        """
        super().__init__(connect_address(protocol, hostname, port), timeout=timeout)


class DummyController(DummyInterface, EventInterface):
    def __init__(self, control_server):
        self._cs = control_server
        self._dev = DummyEthernetInterface(DEV_HOST, DEV_PORT)
        self._dev.connect()

    def info(self) -> str:
        return self._dev.trans("info").decode().strip()

    def get_value(self) -> float:
        return float(self._dev.trans("get_value").decode().strip())

    def handle_event(self, event: Event) -> str:

        _exec_in_thread = False

        def _handle_event(_event):
            _LOGGER.info(f"An event is received, {_event=}")
            _LOGGER.info(f"CM CS active? {is_configuration_manager_active()}")
            time.sleep(5.0)
            _LOGGER.info(f"CM CS active? {is_configuration_manager_active()}")
            _LOGGER.info(f"An event is processed, {_event=}")

        if _exec_in_thread:
            # We execute this function in a daemon thread so the acknowledgment is
            # sent back immediately (the ACK means 'command received and will be
            # executed').

            retry_thread = threading.Thread(target=_handle_event, args=(event,))
            retry_thread.daemon = True
            retry_thread.start()
        else:
            # An alternative to the daemon thread is to create a scheduled task that will be executed
            # after the event is acknowledged.

            self._cs.schedule_task(partial(_handle_event, event))

        return "ACK"


class DummyProtocol(CommandProtocol):

    def __init__(self, control_server: ControlServer):
        super().__init__()
        self.control_server = control_server

        self.device_controller = DummyController(control_server)

        self.load_commands(commands, DummyCommand, DummyController)

        self.build_device_method_lookup_table(self.device_controller)

        self._count = 0

    def get_bind_address(self):
        return bind_address(self.control_server.get_communication_protocol(), self.control_server.get_commanding_port())

    def get_status(self):
        return super().get_status()

    def get_housekeeping(self) -> dict:

        # _LOGGER.debug(f"Executing get_housekeeping function for {self.__class__.__name__}.")

        self._count += 1

        # use the sleep to test the responsiveness of the control server when even this get_housekeeping function takes
        # a lot of time, i.e. up to several minutes in the case of data acquisition devices
        # import time
        # time.sleep(2.0)

        return {
            'timestamp': format_datetime(),
            'COUNT': self._count,
            'PI': 3.14159,  # just to have a constant parameter
            'Random': random.randint(0, 100),  # just to have a variable parameter
        }

    def quit(self):
        _LOGGER.info("Executing 'quit()' on DummyProtocol.")


class DummyControlServer(ControlServer):
    """
    DummyControlServer - Command and monitor dummy device controllers.

    The sever binds to the following ZeroMQ sockets:

    * a REQ-REP socket that can be used as a command server. Any client can connect and
      send a command to the dummy device controller.

    * a PUB-SUP socket that serves as a monitoring server. It will send out status
      information to all the connected clients every HK_DELAY seconds.

    """

    def __init__(self):
        multiprocessing.current_process().name = "dummy_cs"

        super().__init__()

        self.device_protocol = DummyProtocol(self)

        _LOGGER.info(
            f"Binding ZeroMQ socket to {self.device_protocol.get_bind_address()} for {self.__class__.__name__}"
        )

        self.device_protocol.bind(self.dev_ctrl_cmd_sock)

        self.poller.register(self.dev_ctrl_cmd_sock, zmq.POLLIN)

        self.set_hk_delay(cs_settings.HK_DELAY)

        from egse.confman import ConfigurationManagerProxy
        from egse.listener import EVENT_ID

        # The following import is needed because without this import, DummyProxy would be <class '__main__.DummyProxy'>
        # instead of `egse.dummy.DummyProxy` and the ConfigurationManager control server will not be able to de-pickle
        # the register message.
        from egse.dummy import DummyProxy  # noqa

        self.register_as_listener(
            proxy=ConfigurationManagerProxy,
            listener={'name': 'Dummy CS', 'proxy': DummyProxy, 'event_id': EVENT_ID.SETUP}
        )

    def get_communication_protocol(self):
        return 'tcp'

    def get_commanding_port(self):
        return cs_settings.COMMANDING_PORT

    def get_service_port(self):
        return cs_settings.SERVICE_PORT

    def get_monitoring_port(self):
        return cs_settings.MONITORING_PORT

    def get_storage_mnemonic(self):
        return "DUMMY-HK"

    def after_serve(self):
        _LOGGER.debug("After Serve: unregistering Dummy CS")

        from egse.confman import ConfigurationManagerProxy

        self.unregister_as_listener(proxy=ConfigurationManagerProxy, listener={'name': 'Dummy CS'})

    def is_storage_manager_active(self):
        return is_storage_manager_active()

    def store_housekeeping_information(self, data):
        """Send housekeeping information to the Storage manager."""

        store_housekeeping_information(origin=cs_settings.STORAGE_MNEMONIC, data=data)

    def register_to_storage_manager(self) -> None:
        register_to_storage_manager(
            origin=cs_settings.STORAGE_MNEMONIC,
            persistence_class=TYPES["CSV"],
            prep={
                "column_names": list(self.device_protocol.get_housekeeping().keys()),
                "mode": "a",
            }
        )

    def unregister_from_storage_manager(self) -> None:
        unregister_from_storage_manager(origin=cs_settings.STORAGE_MNEMONIC)


@app.command()
def start():
    """Start the dummy control server on localhost."""

    # The following import is needed because without this import, the control server and Proxy will not be able to
    # instantiate classes that are passed in ZeroMQ messages and de-pickled.
    from cgse_dummy.dummy_cs import DummyControlServer  # noqa

    try:
        control_server = DummyControlServer()
        control_server.serve()
    except KeyboardInterrupt:
        print("Shutdown requested...exiting")
    except SystemExit as exit_code:
        print(f"System Exit with code {exit_code}.")
        sys.exit(-1)
    except Exception:  # noqa
        import traceback
        traceback.print_exc(file=sys.stdout)


@app.command()
def stop():
    """Send a quit service command to the dummy control server."""
    with DummyProxy() as dummy:
        _LOGGER.info("Sending quit_server() to Dummy CS.")
        sp = dummy.get_service_proxy()
        sp.quit_server()


if __name__ == '__main__':
    app()



# An example plugin with a dummy device driver
#
import subprocess
import sys
import textwrap
from pathlib import Path

import rich
import typer

dummy = typer.Typer(
    name="dummy",
    help=textwrap.dedent(
        """DUMMY Data Acquisition Unit
        
        This is a simple simulator of a data acquisition device that can be used to monitor
        a number of sensors.         
        """
    ),
    no_args_is_help=True
)


@dummy.command(name="start")
def start_dummy():
    """Start the dummy service, dummy_cs."""
    rich.print("Starting service dummy_cs..")

    out = open(Path('~/.dummy_cs.start.out').expanduser(), 'w')

    subprocess.Popen(
        [sys.executable, '-m', 'cgse_dummy.dummy_cs', 'start'],
        stdout=out, stderr=out, stdin=subprocess.DEVNULL,
        close_fds=True
    )


@dummy.command(name="stop")
def stop_dummy():
    """Stop the dummy service, dummy_cs."""
    rich.print("Terminating service dummy_cs..")

    out = open(Path('~/.dummy_cs.stop.out').expanduser(), 'w')

    subprocess.Popen(
        [sys.executable, '-m', 'cgse_dummy.dummy_cs', 'stop'],
        stdout=out, stderr=out, stdin=subprocess.DEVNULL,
        close_fds=True
    )


@dummy.command(name="status")
def status_dummy():
    """Print the status information from the dummy service, dummy_cs."""
    rich.print("Status information from the dummy service not implemented yet..")


@dummy.command(name="start-sim")
def start_dummy_sim():
    """Start the dummy device Simulator."""
    rich.print("Starting service DUMMY Simulator")

    out = open(Path('~/.dummy_sim.start.out').expanduser(), 'w')

    subprocess.Popen(
        [sys.executable, '-m', 'cgse_dummy.dummy_sim', 'start'],
        stdout=out, stderr=out, stdin=subprocess.DEVNULL,
        close_fds=True
    )


@dummy.command(name="stop-sim")
def stop_dummy_sim():
    """Stop the dummy device Simulator."""
    rich.print("Terminating the DUMMY simulator.")

    out = open(Path('~/.dummy_sim.stop.out').expanduser(), 'w')

    subprocess.Popen(
        [sys.executable, '-m', 'cgse_dummy.dummy_sim', 'stop'],
        stdout=out, stderr=out, stdin=subprocess.DEVNULL,
        close_fds=True
    )


@dummy.command(name="status-sim")
def status_dummy_sim():
    """Print status information on the dummy device simulator."""

    proc = subprocess.Popen(
        [sys.executable, '-m', 'cgse_dummy.dummy_sim', 'status'],
        stdout=subprocess.PIPE, stderr=subprocess.PIPE, stdin=subprocess.DEVNULL
    )

    stdout, stderr = proc.communicate()

    rich.print(stdout.decode(), end='')
    if stderr:
        rich.print(stderr.decode())


if __name__ == '__main__':
    dummy()



"""
The dummy device is a virtual device that is developed as a demonstration of how an external package, that delivers a
device interface for the CGSE, can be implemented.

The simulator listens on the Ethernet socket port number 5555, unless another port number is specified in the
Settings file under the section 'DUMMY DEVICE'.

The following commands are implemented:

- *IDN? — request identification of the device

"""

import contextlib
import datetime
import logging
import re
import select
import socket
import sys

import typer
from egse.device import DeviceConnectionError
from egse.settings import Settings
from egse.system import SignalCatcher
from egse.randomwalk import RandomWalk


logging.basicConfig(level=logging.INFO)

_LOGGER = logging.getLogger("cgse_dummy.dummy.sim")
_VERSION = "0.0.1"

DEVICE_SETTINGS = Settings.load("DUMMY DEVICE")
CS_SETTINGS = Settings.load("DUMMY CS")

try:
    HOSTNAME = CS_SETTINGS.HOSTNAME
except AttributeError:
    HOSTNAME = "localhost"

try:
    PORT = CS_SETTINGS.PORT
except AttributeError:
    PORT = 5555

device_time = datetime.datetime.now(datetime.timezone.utc)
reference_time = device_time
error_msg = ""

sensor_1 = RandomWalk()

app = typer.Typer(help=f"{DEVICE_SETTINGS.BRAND} {DEVICE_SETTINGS.MODEL} Simulator")


def create_datetime(year, month, day, hour, minute, second):
    global device_time, reference_time
    device_time = datetime.datetime(year, month, day, hour, minute, second, tzinfo=datetime.timezone.utc)
    reference_time = datetime.datetime.now(datetime.timezone.utc)


def nothing():
    return None


def set_time(year, month, day, hour, minute, second):
    print(f"TIME {year}, {month}, {day}, {hour}, {minute}, {second}")
    create_datetime(int(year), int(month), int(day), int(hour), int(minute), int(second))


def get_time():
    current_device_time = device_time + (datetime.datetime.now(datetime.timezone.utc) - reference_time)
    msg = current_device_time.strftime("%a %b %d %H:%M:%S %Y")
    print(f":SYST:TIME? {msg = }")
    return msg


def beep(a, b):
    print(f"BEEP {a=}, {b=}")


def reset():
    print("RESET")


def clear():
    print("CLEAR")


def get_value():
    return next(sensor_1)


COMMAND_ACTIONS_RESPONSES = {
    "*IDN?": (None, f"{DEVICE_SETTINGS.BRAND}, MODEL {DEVICE_SETTINGS.MODEL}, SIMULATOR"),
    "*RST": (reset, None),
    "*CLS": (clear, None),

    "info": (None, f"{DEVICE_SETTINGS.BRAND}, MODEL {DEVICE_SETTINGS.MODEL}, SIMULATOR"),
    "get_value": (None, get_value),
}


COMMAND_PATTERNS_ACTIONS_RESPONSES = {
    r":?\*RST": (reset, None),
    r":?SYST(?:em)*:TIME (\d+), (\d+), (\d+), (\d+), (\d+), (\d+)": (set_time, None),
    r":?SYST(?:em)*:TIME\? 1": (nothing, get_time),
    r":?SYST(?:em)*:BEEP(?:er)* (\d+), (\d+(?:\.\d+)?)": (beep, None),
}


def process_command(command_string: str) -> str:

    global COMMAND_ACTIONS_RESPONSES
    global COMMAND_PATTERNS_ACTIONS_RESPONSES

    # LOGGER.debug(f"{command_string=}")

    try:
        action, response = COMMAND_ACTIONS_RESPONSES[command_string]
        action and action()
        if error_msg:
            return error_msg
        else:
            return response if isinstance(response, str) else response()
    except KeyError:
        # try to match with a value
        for key, value in COMMAND_PATTERNS_ACTIONS_RESPONSES.items():
            if match := re.match(key, command_string):
                # LOGGER.debug(f"{match=}, {match.groups()}")
                action, response = value
                # LOGGER.debug(f"{action=}, {response=}")
                action and action(*match.groups())
                return error_msg or (response if isinstance(response, str) or response is None else response())
        return f"ERROR: unknown command string: {command_string}"


def run_simulator():
    """
    Raises:
        OSError: when the simulator is already running.
    """
    global error_msg

    _LOGGER.info(f"Starting the {DEVICE_SETTINGS.MODEL} Simulator")

    killer = SignalCatcher()

    timeout = 2.0

    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.bind((HOSTNAME, PORT))
        sock.listen()
        sock.settimeout(timeout)
        while True:
            while True:
                with contextlib.suppress(socket.timeout):
                    conn, addr = sock.accept()
                    break
                if killer.term_signal_received:
                    return
            with conn:
                _LOGGER.info(f'Accepted connection from {addr}')
                conn.sendall(f'This is {DEVICE_SETTINGS.BRAND} {DEVICE_SETTINGS.MODEL} {_VERSION}.sim'.encode())
                try:
                    error_msg = ""
                    while True:

                        read_sockets, _, _ = select.select([conn], [], [], timeout)

                        if conn in read_sockets:
                            data = conn.recv(4096).decode().rstrip()
                            _LOGGER.debug(f"{data = }")
                            # Now that we use `select` I don't think the following will ever be true
                            # if not data:
                            #     _LOGGER.info("Client closed connection, accepting new connection...")
                            #     break
                            if data.strip() == "STOP":
                                _LOGGER.info("Client requested to terminate...")
                                sock.close()
                                return
                            for cmd in data.split(';'):
                                response = process_command(cmd.strip())
                                if response is not None:
                                    response_b = f"{response}\n".encode()
                                    _LOGGER.debug(f"write: {response_b = }")
                                    conn.sendall(response_b)

                        if killer.term_signal_received:
                            _LOGGER.info("Terminating...")
                            sock.close()
                            return
                        if killer.user_signal_received:
                            if killer.signal_name == "SIGUSR1":
                                _LOGGER.info("SIGUSR1 is not supported by this simulator")
                            if killer.signal_name == "SIGUSR2":
                                _LOGGER.info("SIGUSR2 is not supported by this simulator")
                            killer.clear()

                except ConnectionResetError as exc:
                    _LOGGER.info(f"ConnectionResetError: {exc}")
                except Exception as exc:
                    _LOGGER.info(f"{exc.__class__.__name__} caught: {exc.args}")


def send_request(cmd: str) -> bytes:

    from cgse_dummy.dummy_devif import DummyEthernetInterface

    daq_dev = DummyEthernetInterface(hostname=HOSTNAME, port=PORT)
    daq_dev.connect()

    response = daq_dev.query(cmd)

    daq_dev.disconnect()

    return response


@app.command()
def start():
    try:
        run_simulator()
    except OSError as exc:
        print(f"ERROR: Caught {exc.__class__.__name__}: {exc}", file=sys.stderr)


@app.command()
def status():
    try:
        response = send_request("*IDN?")
        print(f"{response.decode().rstrip()}")
    except DeviceConnectionError as exc:
        print(f"ERROR: Caught {exc.__class__.__name__}: {exc}", file=sys.stderr)


@app.command()
def stop():
    try:
        response = send_request("STOP")
        print(f"{response.decode().rstrip()}")
    except DeviceConnectionError as exc:
        print(f"ERROR: Caught {exc.__class__.__name__}: {exc}", file=sys.stderr)


if __name__ == '__main__':
    app()
